defmodule Amethyst.GameCoordinator do
  use GenServer

  @moduledoc """
  The game coordinator is responsible for keeping track of active
  instances of each game and can create new ones on demand.
  """

  def start_link(initial) when is_list(initial) do
    GenServer.start_link(__MODULE__, initial, name: {:global, __MODULE__})
  end

  @impl true
  def init(initial) do
    {:ok, initial}
  end

  @impl true
  def handle_call({:find, type}, _from, state) do
    {pid, state} = _find(type, state)
    {:reply, pid, state}
  end

  @impl true
  def handle_call({:create, type}, _from, state) do
    {pid, state} = _create(type, state)
    {:reply, pid, state}
  end

  @impl true
  def handle_call({:find_or_create, type}, from, state) do
    {pid, state} = _find(type, state)
    case pid do
      nil -> handle_call({:create, type}, from, state)
      some -> {:reply, some, state}
    end
  end

  defp _create(type, games) do
    state = type.instantiate()
    pid = spawn(type, :start, [state])
    games = [{type, pid, []} | games]
    {pid, games}
  end

  defp _find(type, games) do
    alive_games = games |> Enum.filter(fn {_mod, pid, _opts} -> Process.alive?(pid) end)
    # TODO: Here we should have some additional filtering for specifically joinable games
    existing = alive_games |> Enum.filter(fn {mod, _pid, _opts} -> mod == type end)
    if length(existing) > 0 do
      [{_mod, pid, _} | _] = existing
      {pid, alive_games}
    else
      {nil, alive_games}
    end
  end

  def create(type) do
    GenServer.call({:global, __MODULE__}, {:create, type})
  end
  def find(type) do
    GenServer.call({:global, __MODULE__}, {:find, type})
  end
  def find_or_create(type) do
    GenServer.call({:global, __MODULE__}, {:find_or_create, type})
  end
end
